@@ -1,6 +1,6 @@
 /*
  * LibSylph -- E-Mail client library
- * Copyright (C) 1999-2010 Hiroyuki Yamamoto
+ * Copyright (C) 1999-2012 Hiroyuki Yamamoto
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -47,6 +47,10 @@ static PrefParam param[] = {
 	 P_BOOL},
 	{"newmsg_notify_command", NULL, &prefs_common.newmsg_notify_cmd,
 	 P_STRING},
+	{"enable_newmsg_notify_sound", "FALSE",
+	 &prefs_common.enable_newmsg_notify_sound, P_BOOL},
+	{"newmsg_notify_sound", NULL, &prefs_common.newmsg_notify_sound,
+	 P_STRING},
 
 	{"inc_local", "FALSE", &prefs_common.inc_local, P_BOOL},
 	{"filter_on_inc_local", "TRUE", &prefs_common.filter_on_inc, P_BOOL},
@@ -57,6 +61,7 @@ static PrefParam param[] = {
 	{"filter_sent_message", "FALSE", &prefs_common.filter_sent, P_BOOL},
 	{"recipients_autoreg", "TRUE", &prefs_common.recipients_autoreg,
 	 P_BOOL},
+	{"show_send_dialog", "TRUE", &prefs_common.show_send_dialog, P_BOOL},
 
 	{"encoding_method", "0", &prefs_common.encoding_method, P_ENUM},
 	{"mime_filename_encoding_method", "0",
@@ -304,6 +309,10 @@ static PrefParam param[] = {
 	{"addressbook_y", "32", &prefs_common.addressbook_y, P_INT},
 	{"addressbook_width", "620", &prefs_common.addressbook_width, P_INT},
 	{"addressbook_height", "360", &prefs_common.addressbook_height, P_INT},
+	{"addressbook_folder_width", "190", &prefs_common.addressbook_folder_width, P_INT},
+	{"addressbook_col_name", "164", &prefs_common.addressbook_col_name, P_INT},
+	{"addressbook_col_addr", "156", &prefs_common.addressbook_col_addr, P_INT},
+	{"addressbook_col_rem", "100", &prefs_common.addressbook_col_rem, P_INT},
 
 	/* Message */
 	{"enable_color", "TRUE", &prefs_common.enable_color, P_BOOL},
@@ -319,6 +328,7 @@ static PrefParam param[] = {
 	{"convert_mb_alnum", "FALSE", &prefs_common.conv_mb_alnum, P_BOOL},
 	{"display_header_pane", "TRUE", &prefs_common.display_header_pane,
 	 P_BOOL},
+	{"show_attach_tab", "FALSE", &prefs_common.show_attach_tab, P_BOOL},
 	{"display_header", "TRUE", &prefs_common.display_header, P_BOOL},
 	{"render_html", "TRUE", &prefs_common.render_html, P_BOOL},
 	{"html_only_as_attach", "FALSE", &prefs_common.html_only_as_attach,
@@ -356,11 +366,11 @@ static PrefParam param[] = {
 	/* Junk mail */
 	{"enable_junk", "FALSE", &prefs_common.enable_junk, P_BOOL},
 #ifdef G_OS_WIN32
-	{"junk_learn_command", "bsfilterw -C -s -u",
+	{"junk_learn_command", "sylfilter -j",
 	 &prefs_common.junk_learncmd, P_STRING},
-	{"nojunk_learn_command", "bsfilterw -c -S -u",
+	{"nojunk_learn_command", "sylfilter -c",
 	 &prefs_common.nojunk_learncmd, P_STRING},
-	{"junk_classify_command", "bsfilterw",
+	{"junk_classify_command", "sylfilter",
 	 &prefs_common.junk_classify_cmd, P_STRING},
 #else
 	{"junk_learn_command", "bogofilter -N -s -I",
@@ -398,10 +408,12 @@ static PrefParam param[] = {
 	{"separate_folder", "FALSE", &prefs_common.sep_folder, P_BOOL},
 	{"separate_message", "FALSE", &prefs_common.sep_msg, P_BOOL},
 
-	{"always_show_message_when_selected", "FALSE",
+	{"always_show_message_when_selected", "TRUE",
 	 &prefs_common.always_show_msg, P_BOOL},
 	{"open_unread_on_enter", "FALSE", &prefs_common.open_unread_on_enter,
 	 P_BOOL},
+	{"always_mark_read_on_show_msg", "FALSE",
+	 &prefs_common.always_mark_read_on_show_msg, P_BOOL},
 	{"remember_last_selected", "FALSE",
 	 &prefs_common.remember_last_selected, P_BOOL},
 	{"mark_as_read_on_new_window", "FALSE",
@@ -409,6 +421,8 @@ static PrefParam param[] = {
 	{"open_inbox_on_inc", "FALSE", &prefs_common.open_inbox_on_inc, P_BOOL},
 	{"open_inbox_on_startup", "FALSE", &prefs_common.open_inbox_on_startup,
 	 P_BOOL},
+	{"change_account_on_folder_selection", "TRUE",
+	 &prefs_common.change_account_on_folder_sel, P_BOOL},
 	{"immediate_execution", "TRUE", &prefs_common.immediate_exec, P_BOOL},
 
 #ifdef G_OS_WIN32
@@ -437,6 +451,10 @@ static PrefParam param[] = {
 	 P_BOOL},
 	{"enable_address_completion", "TRUE",
 	 &prefs_common.enable_address_completion, P_BOOL},
+	{"fullauto_completion_mode", "TRUE",
+	 &prefs_common.fullauto_completion_mode, P_BOOL},
+	{"always_add_address_only", "FALSE",
+	 &prefs_common.always_add_address_only, P_BOOL},
 
 	{"confirm_on_exit", "FALSE", &prefs_common.confirm_on_exit, P_BOOL},
 	{"clean_trash_on_exit", "FALSE", &prefs_common.clean_on_exit, P_BOOL},
@@ -484,6 +502,11 @@ static PrefParam param[] = {
 	 P_BOOL},
 	{"io_timeout_secs", "60", &prefs_common.io_timeout_secs, P_INT},
 
+	/* File selector */
+	{"filesel_prev_open_dir", NULL, &prefs_common.prev_open_dir, P_STRING},
+	{"filesel_prev_save_dir", NULL, &prefs_common.prev_save_dir, P_STRING},
+	{"filesel_prev_folder_dir", NULL, &prefs_common.prev_folder_dir, P_STRING},
+
 	{NULL, NULL, NULL, P_OTHER}
 };
 
@@ -512,8 +535,6 @@ void prefs_common_read_config(void)
 
 	prefs_common.online_mode = TRUE;
 
-	prefs_common_junk_filter_list_set();
-
 	path = g_strconcat(get_rc_dir(), G_DIR_SEPARATOR_S, COMMAND_HISTORY,
 			   NULL);
 	if ((fp = g_fopen(path, "rb")) == NULL) {
@@ -560,64 +581,8 @@ void prefs_common_write_config(void)
 	g_free(path);
 }
 
-static FilterRule *prefs_common_junk_filter_rule_create(gboolean is_manual)
-{
-	FilterRule *rule;
-	FilterCond *cond;
-	FilterAction *action;
-	GSList *cond_list = NULL, *action_list = NULL;
-
-	cond = filter_cond_new(FLT_COND_CMD_TEST, 0, 0, NULL,
-			       prefs_common.junk_classify_cmd);
-	cond_list = g_slist_append(NULL, cond);
-	if (prefs_common.delete_junk_on_recv && !is_manual) {
-		action = filter_action_new(FLT_ACTION_COPY,
-					   prefs_common.junk_folder);
-		action_list = g_slist_append(NULL, action);
-		action = filter_action_new(FLT_ACTION_DELETE, NULL);
-		action_list = g_slist_append(action_list, action);
-	} else {
-		action = filter_action_new(FLT_ACTION_MOVE,
-					   prefs_common.junk_folder);
-		action_list = g_slist_append(NULL, action);
-	}
-
-	if (prefs_common.mark_junk_as_read) {
-		action = filter_action_new(FLT_ACTION_MARK_READ, NULL);
-		action_list = g_slist_append(action_list, action);
-	}
-
-	if (is_manual)
-		rule = filter_rule_new(_("Junk mail filter (manual)"), FLT_OR,
-	 			       cond_list, action_list);
-	else
-		rule = filter_rule_new(_("Junk mail filter"), FLT_OR,
-				       cond_list, action_list);
-
-	return rule;
-}
-
 void prefs_common_junk_filter_list_set(void)
 {
-	FilterRule *rule;
-
-	if (prefs_common.junk_fltlist) {
-		filter_rule_list_free(prefs_common.junk_fltlist);
-		prefs_common.junk_fltlist = NULL;
-	}
-	if (prefs_common.manual_junk_fltlist) {
-		filter_rule_list_free(prefs_common.manual_junk_fltlist);
-		prefs_common.manual_junk_fltlist = NULL;
-	}
-
-	if (!prefs_common.junk_classify_cmd || !prefs_common.junk_folder)
-		return;
-
-	rule = prefs_common_junk_filter_rule_create(FALSE);
-	prefs_common.junk_fltlist = g_slist_append(NULL, rule);
-
-	rule = prefs_common_junk_filter_rule_create(TRUE);
-	prefs_common.manual_junk_fltlist = g_slist_append(NULL, rule);
 }
 
 void prefs_common_junk_folder_rename_path(const gchar *old_path,
@@ -649,6 +614,5 @@ void prefs_common_junk_folder_rename_path(const gchar *old_path,
 			    dest_path);
 		g_free(prefs_common.junk_folder);
 		prefs_common.junk_folder = dest_path;
-		prefs_common_junk_filter_list_set();
 	}
 }
