@@ -1,6 +1,6 @@
 /*
  * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
- * Copyright (C) 1999-2010 Hiroyuki Yamamoto
+ * Copyright (C) 1999-2012 Hiroyuki Yamamoto
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -64,7 +64,6 @@
 #include "filter.h"
 #include "send_message.h"
 #include "inc.h"
-#include "import.h"
 #include "manage_window.h"
 #include "alertpanel.h"
 #include "inputdialog.h"
@@ -113,6 +112,7 @@ static gboolean init_console_done = FALSE;
 static gint lock_socket = -1;
 static gint lock_socket_tag = 0;
 static GIOChannel *lock_ch = NULL;
+static gchar *instance_id = NULL;
 
 #if USE_THREADS
 static GThread *main_thread;
@@ -132,6 +132,8 @@ static struct RemoteCmd {
 	gchar *open_msg;
 	gboolean configdir;
 	gboolean exit;
+	gboolean restart;
+	gchar *argv0;
 #ifdef G_OS_WIN32
 	gushort ipcport;
 #endif
@@ -150,6 +152,10 @@ static struct RemoteCmd {
 			  mainwin->mainwin_cid); \
 }
 
+#if defined(G_OS_WIN32) || defined(__APPLE__)
+static void fix_font_setting		(void);
+#endif
+
 static void parse_cmd_opt		(int		 argc,
 					 char		*argv[]);
 
@@ -169,7 +175,10 @@ static gboolean lock_socket_input_cb	(GIOChannel	*source,
 					 gpointer	 data);
 
 static void remote_command_exec		(void);
+
+#if !defined(G_OS_WIN32) && !defined(__APPLE__)
 static void migrate_old_config		(void);
+#endif
 
 static void open_compose_new		(const gchar	*address,
 					 GPtrArray	*attach_files);
@@ -200,7 +209,7 @@ static void send_queue			(void);
 
 static void load_cb(GObject *obj, GModule *module, gpointer data)
 {
-	g_print("load_cb: %p (%s), %p\n", module, module ? g_module_name(module) : "(null)", data);
+	debug_print("load_cb: %p (%s), %p\n", module, module ? g_module_name(module) : "(null)", data);
 }
 
 int main(int argc, char *argv[])
@@ -213,6 +222,7 @@ int main(int argc, char *argv[])
 #endif
 	GObject *syl_app;
 	PrefsAccount *new_account = NULL;
+	gboolean first_run = FALSE;
 
 	app_init();
 	parse_cmd_opt(argc, argv);
@@ -252,6 +262,7 @@ int main(int argc, char *argv[])
 	set_progress_func(main_window_progress_show);
 	set_input_query_password_func(input_dialog_query_password);
 #if USE_SSL
+	ssl_init();
 	ssl_set_verify_func(ssl_manager_verify_cert);
 #endif
 
@@ -264,32 +275,14 @@ int main(int argc, char *argv[])
 	prefs_display_header_read_config();
 	colorlabel_read_config();
 
-#ifdef G_OS_WIN32
-	{
-		gchar *path;
-		path = g_strconcat(get_rc_dir(), G_DIR_SEPARATOR_S, COMMON_RC,
-				   NULL);
-		if (!is_file_exist(path) && conv_is_ja_locale()) {
-			const gchar *str;
+	prefs_common.user_agent_str = g_strdup_printf
+		("%s (GTK+ %d.%d.%d; %s)",
+		 prog_version,
+		 gtk_major_version, gtk_minor_version, gtk_micro_version,
+		 TARGET_ALIAS);
 
-			debug_print("fixing prefs_common.textfont setting\n");
-			str = "MS Gothic 12";
-			if (!gtkut_font_can_load(str)) {
-				debug_print("font '%s' load failed\n", str);
-				str = "\xef\xbc\xad\xef\xbc\xb3 \xe3\x82\xb4\xe3\x82\xb7\xe3\x83\x83\xe3\x82\xaf 12";
-				if (!gtkut_font_can_load(str)) {
-					debug_print("font '%s' load failed\n", str);
-					str = NULL;
-				}
-			}
-			if (str) {
-				debug_print("font '%s' load ok\n", str);
-				g_free(prefs_common.textfont);
-				prefs_common.textfont = g_strdup(str);
-			}
-		}
-		g_free(path);
-	}
+#if defined(G_OS_WIN32) || defined(__APPLE__)
+	fix_font_setting();
 #endif
 
 	gtkut_stock_button_set_set_reverse(!prefs_common.comply_gnome_hig);
@@ -331,6 +324,7 @@ int main(int argc, char *argv[])
 	main_window_reflect_prefs_all();
 
 	if (folder_read_list() < 0) {
+		first_run = TRUE;
 		setup_mailbox();
 		folder_write_list();
 	}
@@ -357,6 +351,11 @@ int main(int argc, char *argv[])
 
 	g_signal_emit_by_name(syl_app, "init-done");
 
+	if (first_run) {
+		setup_import_data();
+		setup_import_addressbook();
+	}
+
 	remote_command_exec();
 
 #if USE_UPDATE_CHECK
@@ -442,6 +441,54 @@ static void read_ini_file(void)
 		cmd.configdir = TRUE;
 	}
 }
+#endif /* G_OS_WIN32 */
+
+#if defined(G_OS_WIN32) || defined(__APPLE__)
+static void fix_font_setting(void)
+{
+	const gchar *str = NULL;
+
+	if (!conv_is_ja_locale())
+		return;
+
+	if (prefs_common.textfont &&
+	    strcmp(prefs_common.textfont, DEFAULT_MESSAGE_FONT) != 0) {
+		if (gtkut_font_can_load(prefs_common.textfont)) {
+			debug_print("font '%s' load ok\n", prefs_common.textfont);
+			return;
+		}
+		debug_print("font '%s' load failed\n", prefs_common.textfont);
+	}
+
+	debug_print("fixing prefs_common.textfont setting\n");
+
+#ifdef G_OS_WIN32
+	str = "MS Gothic 12";
+#else /* __APPLE__ */
+	str = "Hiragino Kaku Gothic Pro Light 13";
+#endif
+
+	if (!gtkut_font_can_load(str)) {
+#ifdef G_OS_WIN32
+		debug_print("font '%s' load failed\n", str);
+		str = "\xef\xbc\xad\xef\xbc\xb3 \xe3\x82\xb4\xe3\x82\xb7\xe3\x83\x83\xe3\x82\xaf 12";
+		if (!gtkut_font_can_load(str)) {
+			debug_print("font '%s' load failed\n", str);
+			str = NULL;
+		}
+#else /* __APPLE__ */
+		debug_print("font '%s' load failed\n", str);
+		str = NULL;
+#endif
+	}
+
+	if (str) {
+		debug_print("font '%s' load ok\n", str);
+		g_free(prefs_common.textfont);
+		prefs_common.textfont = g_strdup(str);
+	} else
+		g_warning("failed to load text font!");
+}
 #endif
 
 static void parse_cmd_opt(int argc, char *argv[])
@@ -551,6 +598,12 @@ static void parse_cmd_opt(int argc, char *argv[])
 				i++;
 			}
 #endif
+		} else if (!strncmp(argv[i], "--instance-id", 13)) {
+			if (argv[i + 1]) {
+				instance_id = g_locale_to_utf8
+					(argv[i + 1], -1, NULL, NULL, NULL);
+				i++;
+			}
 		} else if (!strncmp(argv[i], "--exit", 6)) {
 			cmd.exit = TRUE;
 		} else if (!strncmp(argv[i], "--help", 6)) {
@@ -594,6 +647,10 @@ static void parse_cmd_opt(int argc, char *argv[])
 		cmd.compose = TRUE;
 		cmd.compose_mailto = NULL;
 	}
+
+	cmd.argv0 = g_locale_to_utf8(argv[0], -1, NULL, NULL, NULL);
+	if (!cmd.argv0)
+		cmd.argv0 = g_strdup(argv[0]);
 }
 
 static gint get_queued_message_num(void)
@@ -692,7 +749,7 @@ static void parse_gtkrc_files(void)
 
 static void setup_rc_dir(void)
 {
-#ifndef G_OS_WIN32
+#if !defined(G_OS_WIN32) && !defined(__APPLE__)
 	CHDIR_EXIT_IF_FAIL(get_home_dir(), 1);
 
 	/* backup if old rc file exists */
@@ -732,11 +789,55 @@ static void setup_rc_dir(void)
 		if (is_dir_exist(OLD_RC_DIR))
 			migrate_old_config();
 	}
-#endif /* !G_OS_WIN32 */
+#endif /* !G_OS_WIN32 && !__APPLE__ */
 
 	syl_setup_rc_dir();
 }
 
+static void app_restart(void)
+{
+	gchar *cmdline;
+	GError *error = NULL;
+#ifdef G_OS_WIN32
+	if (cmd.configdir) {
+		cmdline = g_strdup_printf("\"%s\"%s --configdir \"%s\" --ipcport %d",
+					  cmd.argv0,
+					  get_debug_mode() ? " --debug" : "",
+					  get_rc_dir(),
+					  cmd.ipcport);
+	} else {
+		cmdline = g_strdup_printf("\"%s\"%s --ipcport %d",
+					  cmd.argv0,
+					  get_debug_mode() ? " --debug" : "",
+					  cmd.ipcport);
+	}
+#else
+	if (cmd.configdir) {
+		cmdline = g_strdup_printf("\"%s\"%s --configdir \"%s\"",
+					  cmd.argv0,
+					  get_debug_mode() ? " --debug" : "",
+					  get_rc_dir());
+	} else {
+		cmdline = g_strdup_printf("\"%s\"%s",
+					  cmd.argv0,
+					  get_debug_mode() ? " --debug" : "");
+	}
+#endif
+	if (!g_spawn_command_line_async(cmdline, &error)) {
+		alertpanel_error("restart failed\n'%s'\n%s", cmdline, error->message);
+		g_error_free(error);
+	}
+	g_free(cmdline);
+}
+
+void app_will_restart(gboolean force)
+{
+	cmd.restart = TRUE;
+	app_will_exit(force);
+	/* canceled */
+	cmd.restart = FALSE;
+}
+
 void app_will_exit(gboolean force)
 {
 	MainWindow *mainwin;
@@ -773,6 +874,8 @@ void app_will_exit(gboolean force)
 		manage_window_focus_in(mainwin->window, NULL, NULL);
 	}
 
+	if (force)
+		g_signal_emit_by_name(syl_app_get(), "app-force-exit");
 	g_signal_emit_by_name(syl_app_get(), "app-exit");
 
 	inc_autocheck_timer_remove();
@@ -804,14 +907,29 @@ void app_will_exit(gboolean force)
 	g_free(filename);
 
 	/* remove temporary files, close log file, socket cleanup */
+#if USE_SSL
+	ssl_done();
+#endif
 	syl_cleanup();
 	lock_socket_remove();
 
+#ifdef USE_UPDATE_CHECK_PLUGIN
+#ifdef G_OS_WIN32
+	cur = gtk_window_list_toplevels();
+	g_list_foreach(cur, (GFunc)gtk_widget_hide, NULL);
+	g_list_free(cur);
+	update_check_spawn_plugin_updater();
+#endif
+#endif
+
 	cleanup_console();
 
 	if (gtk_main_level() > 0)
 		gtk_main_quit();
 
+	if (cmd.restart)
+		app_restart();
+
 	exit(0);
 }
 
@@ -1096,6 +1214,8 @@ static void register_system_events(void)
 }
 #endif
 
+#define ADD_SYM(sym)	syl_plugin_add_symbol(#sym, sym)
+
 static void plugin_init(void)
 {
 	MainWindow *mainwin;
@@ -1110,19 +1230,32 @@ static void plugin_init(void)
 		return;
 	}
 
-	syl_plugin_add_symbol("prog_version", prog_version);
-	syl_plugin_add_symbol("main_window_lock", main_window_lock);
-	syl_plugin_add_symbol("main_window_unlock", main_window_unlock);
-	syl_plugin_add_symbol("main_window_get", main_window_get);
-	syl_plugin_add_symbol("main_window_popup", main_window_popup);
-	syl_plugin_add_symbol("app_will_exit", app_will_exit);
+	ADD_SYM(prog_version);
+	ADD_SYM(app_will_exit);
+
+	ADD_SYM(main_window_lock);
+	ADD_SYM(main_window_unlock);
+	ADD_SYM(main_window_get);
+	ADD_SYM(main_window_popup);
+
 	syl_plugin_add_symbol("main_window_menu_factory",
 			      mainwin->menu_factory);
 	syl_plugin_add_symbol("main_window_statusbar", mainwin->statusbar);
 
-	syl_plugin_add_symbol("folderview", mainwin->folderview);
-	syl_plugin_add_symbol("folderview_get_selected_item",
-			      folderview_get_selected_item);
+	ADD_SYM(folderview_get);
+	ADD_SYM(folderview_add_sub_widget);
+	ADD_SYM(folderview_select);
+	ADD_SYM(folderview_unselect);
+	ADD_SYM(folderview_select_next_unread);
+	ADD_SYM(folderview_get_selected_item);
+	ADD_SYM(folderview_check_new);
+	ADD_SYM(folderview_check_new_item);
+	ADD_SYM(folderview_check_new_all);
+	ADD_SYM(folderview_update_item);
+	ADD_SYM(folderview_update_item_foreach);
+	ADD_SYM(folderview_update_all_updated);
+	ADD_SYM(folderview_check_new_selected);
+
 	syl_plugin_add_symbol("folderview_mail_popup_factory",
 			      mainwin->folderview->mail_factory);
 	syl_plugin_add_symbol("folderview_imap_popup_factory",
@@ -1131,44 +1264,80 @@ static void plugin_init(void)
 			      mainwin->folderview->news_factory);
 
 	syl_plugin_add_symbol("summaryview", mainwin->summaryview);
-	syl_plugin_add_symbol("summary_select_by_msgnum",
-			      summary_select_by_msgnum);
-	syl_plugin_add_symbol("summary_select_by_msginfo",
-			      summary_select_by_msginfo);
-	syl_plugin_add_symbol("summary_lock", summary_lock);
-	syl_plugin_add_symbol("summary_unlock", summary_unlock);
-	syl_plugin_add_symbol("summary_is_locked", summary_is_locked);
+	syl_plugin_add_symbol("summaryview_popup_factory",
+			      mainwin->summaryview->popupfactory);
 
-	syl_plugin_add_symbol("messageview_create_with_new_window",
-			      messageview_create_with_new_window);
-	syl_plugin_add_symbol("messageview_show", messageview_show);
+	ADD_SYM(summary_select_by_msgnum);
+	ADD_SYM(summary_select_by_msginfo);
+	ADD_SYM(summary_lock);
+	ADD_SYM(summary_unlock);
+	ADD_SYM(summary_is_locked);
+	ADD_SYM(summary_is_read_locked);
+	ADD_SYM(summary_write_lock);
+	ADD_SYM(summary_write_unlock);
+	ADD_SYM(summary_is_write_locked);
+	ADD_SYM(summary_get_current_folder);
+	ADD_SYM(summary_get_selection_type);
+	ADD_SYM(summary_get_selected_msg_list);
+	ADD_SYM(summary_get_msg_list);
+	ADD_SYM(summary_show_queued_msgs);
+	ADD_SYM(summary_redisplay_msg);
+	ADD_SYM(summary_open_msg);
+	ADD_SYM(summary_view_source);
+	ADD_SYM(summary_reedit);
+	ADD_SYM(summary_update_selected_rows);
+	ADD_SYM(summary_update_by_msgnum);
 
-	syl_plugin_add_symbol("compose_new", compose_new);
-	syl_plugin_add_symbol("compose_entry_set", compose_entry_set);
-	syl_plugin_add_symbol("compose_entry_append", compose_entry_append);
-	syl_plugin_add_symbol("compose_entry_get_text", compose_entry_get_text);
-	syl_plugin_add_symbol("compose_lock", compose_lock);
-	syl_plugin_add_symbol("compose_unlock", compose_unlock);
+	ADD_SYM(messageview_create_with_new_window);
+	ADD_SYM(messageview_show);
 
-	syl_plugin_add_symbol("foldersel_folder_sel",
-			      foldersel_folder_sel);
-	syl_plugin_add_symbol("foldersel_folder_sel_full",
-			      foldersel_folder_sel_full);
-	syl_plugin_add_symbol("input_dialog", input_dialog);
-	syl_plugin_add_symbol("input_dialog_with_invisible",
-			      input_dialog_with_invisible);
+	ADD_SYM(compose_new);
+	ADD_SYM(compose_entry_set);
+	ADD_SYM(compose_entry_append);
+	ADD_SYM(compose_entry_get_text);
+	ADD_SYM(compose_lock);
+	ADD_SYM(compose_unlock);
 
-	syl_plugin_add_symbol("manage_window_set_transient",
-			      manage_window_set_transient);
-	syl_plugin_add_symbol("manage_window_signals_connect",
-			      manage_window_signals_connect);
-	syl_plugin_add_symbol("manage_window_get_focus_window",
-			      manage_window_get_focus_window);
+	ADD_SYM(foldersel_folder_sel);
+	ADD_SYM(foldersel_folder_sel_full);
 
-	syl_plugin_add_symbol("inc_mail", inc_mail);
-	syl_plugin_add_symbol("inc_is_active", inc_is_active);
-	syl_plugin_add_symbol("inc_lock", inc_lock);
-	syl_plugin_add_symbol("inc_unlock", inc_unlock);
+	ADD_SYM(input_dialog);
+	ADD_SYM(input_dialog_with_invisible);
+
+	ADD_SYM(manage_window_set_transient);
+	ADD_SYM(manage_window_signals_connect);
+	ADD_SYM(manage_window_get_focus_window);
+
+	ADD_SYM(inc_mail);
+	ADD_SYM(inc_is_active);
+	ADD_SYM(inc_lock);
+	ADD_SYM(inc_unlock);
+
+#if USE_UPDATE_CHECK
+	ADD_SYM(update_check);
+	ADD_SYM(update_check_set_check_url);
+	ADD_SYM(update_check_get_check_url);
+	ADD_SYM(update_check_set_download_url);
+	ADD_SYM(update_check_get_download_url);
+	ADD_SYM(update_check_set_jump_url);
+	ADD_SYM(update_check_get_jump_url);
+#ifdef USE_UPDATE_CHECK_PLUGIN
+	ADD_SYM(update_check_set_check_plugin_url);
+	ADD_SYM(update_check_get_check_plugin_url);
+	ADD_SYM(update_check_set_jump_plugin_url);
+	ADD_SYM(update_check_get_jump_plugin_url);
+#endif /* USE_UPDATE_CHECK_PLUGIN */
+#endif
+
+	ADD_SYM(alertpanel_full);
+	ADD_SYM(alertpanel);
+	ADD_SYM(alertpanel_message);
+	ADD_SYM(alertpanel_message_with_disable);
+
+	ADD_SYM(send_message);
+	ADD_SYM(send_message_queue_all);
+	ADD_SYM(send_message_set_reply_flag);
+	ADD_SYM(send_message_set_forward_flags);
 
 	syl_plugin_signal_connect("plugin-load", G_CALLBACK(load_cb), NULL);
 
@@ -1195,8 +1364,9 @@ static gchar *get_socket_name(void)
 	static gchar *filename = NULL;
 
 	if (filename == NULL) {
-		filename = g_strdup_printf("%s%csylpheed-%d",
+		filename = g_strdup_printf("%s%c%s-%d",
 					   g_get_tmp_dir(), G_DIR_SEPARATOR,
+					   instance_id ? instance_id : "sylpheed",
 #if HAVE_GETUID
 					   getuid());
 #else
@@ -1213,28 +1383,35 @@ static gint prohibit_duplicate_launch(void)
 
 #ifdef G_OS_WIN32
 	HANDLE hmutex;
+	const gchar *ins_id = instance_id ? instance_id : "Sylpheed";
+	gushort port = cmd.ipcport ? cmd.ipcport : REMOTE_CMD_PORT;
 
-	hmutex = CreateMutexA(NULL, FALSE, "Sylpheed");
+	debug_print("prohibit_duplicate_launch: checking mutex: %s\n", ins_id);
+	hmutex = CreateMutexA(NULL, FALSE, ins_id);
 	if (!hmutex) {
-		g_warning("cannot create Mutex\n");
+		g_warning("cannot create Mutex: %s\n", ins_id);
 		return -1;
 	}
 	if (GetLastError() != ERROR_ALREADY_EXISTS) {
-		sock = fd_open_inet(cmd.ipcport ? cmd.ipcport : REMOTE_CMD_PORT);
+		debug_print("prohibit_duplicate_launch: creating socket: port %d\n", port);
+		sock = fd_open_inet(port);
 		if (sock < 0)
 			return 0;
 		return sock;
 	}
 
-	sock = fd_connect_inet(cmd.ipcport ? cmd.ipcport : REMOTE_CMD_PORT);
+	debug_print("prohibit_duplicate_launch: connecting to socket: port %d\n", port);
+	sock = fd_connect_inet(port);
 	if (sock < 0)
 		return -1;
 #else
 	gchar *path;
 
 	path = get_socket_name();
+	debug_print("prohibit_duplicate_launch: checking socket: %s\n", path);
 	sock = fd_connect_unix(path);
 	if (sock < 0) {
+		debug_print("prohibit_duplicate_launch: creating socket: %s\n", path);
 		g_unlink(path);
 		return fd_open_unix(path);
 	}
@@ -1346,6 +1523,7 @@ static gint lock_socket_remove(void)
 
 #ifndef G_OS_WIN32
 	filename = get_socket_name();
+	debug_print("lock_socket_remove: removing socket: %s\n", filename);
 	g_unlink(filename);
 #endif
 
@@ -1471,13 +1649,24 @@ static void remote_command_exec(void)
 
 	if (prefs_common.open_inbox_on_startup) {
 		FolderItem *item;
-		item = cur_account && cur_account->inbox
-			? folder_find_item_from_identifier(cur_account->inbox)
+		PrefsAccount *ac;
+
+		ac = account_get_default();
+		if (!ac)
+			ac = cur_account;
+		item = ac && ac->inbox
+			? folder_find_item_from_identifier(ac->inbox)
 			: folder_get_default_inbox();
 		folderview_select(mainwin->folderview, item);
 	}
 
 	if (!gtkut_window_modal_exist()) {
+		if (cmd.compose)
+			open_compose_new(cmd.compose_mailto, cmd.attach_files);
+
+		if (cmd.open_msg)
+			open_message(cmd.open_msg);
+
 		if (cmd.receive_all)
 			inc_all_account_mail(mainwin, FALSE);
 		else if (prefs_common.chk_on_startup)
@@ -1485,14 +1674,8 @@ static void remote_command_exec(void)
 		else if (cmd.receive)
 			inc_mail(mainwin);
 
-		if (cmd.compose)
-			open_compose_new(cmd.compose_mailto, cmd.attach_files);
-
 		if (cmd.send)
 			send_queue();
-
-		if (cmd.open_msg)
-			open_message(cmd.open_msg);
 	}
 
 	if (cmd.attach_files) {
@@ -1517,6 +1700,7 @@ static void remote_command_exec(void)
 	}
 }
 
+#if !defined(G_OS_WIN32) && !defined(__APPLE__)
 static void migrate_old_config(void)
 {
 	GDir *dir;
@@ -1594,6 +1778,7 @@ static void migrate_old_config(void)
 	g_pattern_spec_free(pspec);
 	g_dir_close(dir);
 }
+#endif /* !G_OS_WIN32 && !__APPLE__ */
 
 static void open_compose_new(const gchar *address, GPtrArray *attach_files)
 {
