@@ -1,6 +1,6 @@
 /*
  * Sylpheed -- a GTK+ based, lightweight, and fast e-mail client
- * Copyright (C) 1999-2007 Hiroyuki Yamamoto
+ * Copyright (C) 1999-2012 Hiroyuki Yamamoto
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -56,6 +56,8 @@
 #include "displayheader.h"
 #include "filesel.h"
 #include "alertpanel.h"
+#include "menu.h"
+#include "plugin.h"
 
 typedef struct _RemoteURI	RemoteURI;
 
@@ -142,6 +144,8 @@ static GdkCursor *hand_cursor = NULL;
 static GdkCursor *regular_cursor = NULL;
 
 
+static void textview_part_menu_create	(TextView	*textview);
+
 static void textview_add_part		(TextView	*textview,
 					 MimeInfo	*mimeinfo,
 					 FILE		*fp);
@@ -174,6 +178,10 @@ static GPtrArray *textview_scan_header	(TextView	*textview,
 static void textview_show_header	(TextView	*textview,
 					 GPtrArray	*headers);
 
+static void textview_insert_border	(TextView	*textview,
+					 GtkTextIter	*iter,
+					 gint		 padding);
+
 static gboolean textview_key_pressed		(GtkWidget	*widget,
 						 GdkEventKey	*event,
 						 TextView	*textview);
@@ -196,6 +204,9 @@ static void textview_populate_popup		(GtkWidget	*widget,
 static void textview_popup_menu_activate_open_uri_cb
 						(GtkMenuItem	*menuitem,
 						 gpointer	 data);
+static void textview_popup_menu_activate_reply_cb
+						(GtkMenuItem	*menuitem,
+						 gpointer	 data);
 static void textview_popup_menu_activate_add_address_cb
 						(GtkMenuItem	*menuitem,
 						 gpointer	 data);
@@ -305,6 +316,8 @@ TextView *textview_create(void)
 	textview->body_pos         = 0;
 	textview->show_all_headers = FALSE;
 
+	textview_part_menu_create(textview);
+
 	return textview;
 }
 
@@ -614,6 +627,276 @@ void textview_show_part(TextView *textview, MimeInfo *mimeinfo, FILE *fp)
 	gtk_text_view_scroll_mark_onscreen(text, mark);
 }
 
+enum {
+	PART_MENU_NONE,
+	PART_MENU_OPEN,
+	PART_MENU_OPEN_WITH,
+	PART_MENU_SAVE_AS,
+	PART_MENU_PRINT,
+	PART_MENU_COPY_FILENAME
+};
+
+static void part_menu_button_position(GtkMenu *menu, gint *x, gint *y,
+				      gboolean *push_in,
+				      gpointer user_data)
+{
+	GtkWidget *widget;
+	GtkRequisition requisition;
+	gint button_xpos, button_ypos;
+	gint xpos, ypos;
+	gint width, height;
+	gint scr_width, scr_height;
+
+	g_return_if_fail(x != NULL && y != NULL);
+
+	widget = GTK_WIDGET(user_data);
+	gtk_widget_get_child_requisition(GTK_WIDGET(menu), &requisition);
+	width = requisition.width;
+	height = requisition.height;
+	gdk_window_get_origin(widget->window, &button_xpos, &button_ypos);
+
+	xpos = button_xpos;
+	ypos = button_ypos + widget->requisition.height;
+
+	scr_width = gdk_screen_width();
+	scr_height = gdk_screen_height();
+
+	if (xpos + width > scr_width)
+		xpos -= (xpos + width) - scr_width;
+	if (ypos + height > scr_height)
+		ypos -= widget->requisition.height + height;
+	if (xpos < 0)
+		xpos = 0;
+	if (ypos < 0)
+		ypos = 0;
+
+	*x = xpos;
+	*y = ypos;
+}
+
+static gboolean textview_part_widget_entered(GtkWidget *widget,
+					     GdkEventCrossing *event,
+					     gpointer data)
+{
+	gtk_widget_set_state(widget, GTK_STATE_PRELIGHT);
+	return FALSE;
+}
+
+static gboolean textview_part_widget_left(GtkWidget *widget,
+					  GdkEventCrossing *event,
+					  gpointer data)
+{
+	gtk_widget_set_state(widget, GTK_STATE_NORMAL);
+	return FALSE;
+}
+
+static gboolean textview_part_widget_button_pressed(GtkWidget *widget,
+						    GdkEventButton *event,
+						    gpointer data)
+{
+	TextView *textview = (TextView *)data;
+	GtkWidget *menu;
+	MimeInfo *mimeinfo;
+	GList *cur;
+
+	if (!event)
+		return FALSE;
+
+	menu = textview->popup_menu;
+	mimeinfo = g_object_get_data(G_OBJECT(widget), "mimeinfo");
+
+	for (cur = GTK_MENU_SHELL(menu)->children; cur != NULL; cur = cur->next) {
+		GtkWidget *menuitem = GTK_WIDGET(cur->data);
+		gint type = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(menuitem), MENU_VAL_ID));
+		if (type == PART_MENU_PRINT) {
+			if (mimeinfo->mime_type == MIME_TEXT ||
+			    mimeinfo->mime_type == MIME_TEXT_HTML ||
+			    mimeinfo->mime_type == MIME_MESSAGE_RFC822)
+				gtk_widget_set_sensitive(menuitem, TRUE);
+			else
+				gtk_widget_set_sensitive(menuitem, FALSE);
+		} else if (type == PART_MENU_COPY_FILENAME) {
+			if (mimeinfo->filename || mimeinfo->name)
+				gtk_widget_set_sensitive(menuitem, TRUE);
+			else
+				gtk_widget_set_sensitive(menuitem, FALSE);
+		}
+	}
+
+	g_object_set_data(G_OBJECT(menu), "part-widget", widget);
+	gtk_menu_popup(GTK_MENU(menu), NULL, NULL, part_menu_button_position, widget, event->button, event->time);
+
+	return TRUE;
+}
+
+static gboolean textview_part_widget_exposed(GtkWidget *widget,
+					     GdkEventExpose *event,
+					     gpointer data)
+{
+	GdkDrawable *drawable;
+	GdkGC *gc;
+	gint w, h;
+
+	w = widget->allocation.width - 1;
+	h = widget->allocation.height - 1;
+	if (w <= 0 || h <= 0)
+		return FALSE;
+
+	drawable = GDK_DRAWABLE(widget->window);
+	gc = widget->style->dark_gc[GTK_WIDGET_STATE(widget)];
+	gdk_gc_set_clip_rectangle(gc, &event->area);
+	gdk_gc_set_line_attributes(gc, 1, GDK_LINE_SOLID, GDK_CAP_NOT_LAST,
+				   GDK_JOIN_MITER);
+	gdk_draw_line(drawable, gc, 1, 0, w, 0);
+	gdk_draw_line(drawable, gc, w, 1, w, h);
+	gdk_draw_line(drawable, gc, w, h, 1, h);
+	gdk_draw_line(drawable, gc, 0, h, 0, 1);
+
+	return FALSE;
+}
+
+static void textview_part_menu_activated(GtkWidget *widget, gpointer data)
+{
+	TextView *textview = (TextView *)data;
+	GtkWidget *menu;
+	GtkWidget *part_widget;
+	gint type;
+	MimeInfo *mimeinfo;
+	MimeView *mimeview = textview->messageview->mimeview;
+	const gchar *filename;
+	GtkClipboard *clipboard;
+
+	menu = gtk_widget_get_parent(widget);
+	part_widget = g_object_get_data(G_OBJECT(menu), "part-widget");
+	mimeinfo = g_object_get_data(G_OBJECT(part_widget), "mimeinfo");
+	if (!mimeinfo)
+		return;
+	type = GPOINTER_TO_INT(g_object_get_data(G_OBJECT(widget), MENU_VAL_ID));
+
+	switch (type) {
+	case PART_MENU_OPEN:
+		mimeview_launch_part(mimeview, mimeinfo);
+		break;
+	case PART_MENU_OPEN_WITH:
+		mimeview_open_part_with(mimeview, mimeinfo);
+		break;
+	case PART_MENU_SAVE_AS:
+		mimeview_save_part_as(mimeview, mimeinfo);
+		break;
+	case PART_MENU_PRINT:
+		mimeview_print_part(mimeview, mimeinfo);
+		break;
+	case PART_MENU_COPY_FILENAME:
+		filename = mimeinfo->filename ? mimeinfo->filename :
+			   mimeinfo->name ? mimeinfo->name : NULL;
+		if (filename) {
+			clipboard = gtk_clipboard_get(GDK_SELECTION_PRIMARY);
+			gtk_clipboard_set_text(clipboard, filename, -1);
+			clipboard = gtk_clipboard_get(GDK_SELECTION_CLIPBOARD);
+			gtk_clipboard_set_text(clipboard, filename, -1);
+		}
+		break;
+	default:
+		break;
+	}
+}
+
+static void textview_part_menu_closed(GtkMenuShell *menu_shell, gpointer data)
+{
+	g_object_set_data(G_OBJECT(menu_shell), "part-widget", NULL);
+}
+
+static void textview_part_menu_create(TextView *textview)
+{
+	GtkWidget *menu;
+	GtkWidget *menuitem;
+
+	if (textview->popup_menu)
+		return;
+
+	menu = gtk_menu_new();
+	textview->popup_menu = menu;
+
+	MENUITEM_ADD_FROM_STOCK(menu, menuitem, GTK_STOCK_OPEN, PART_MENU_OPEN);
+	g_signal_connect(G_OBJECT(menuitem), "activate",
+			 G_CALLBACK(textview_part_menu_activated), textview);
+	MENUITEM_ADD_WITH_MNEMONIC(menu, menuitem, _("Open _with..."), PART_MENU_OPEN_WITH);
+	g_signal_connect(G_OBJECT(menuitem), "activate",
+			 G_CALLBACK(textview_part_menu_activated), textview);
+	MENUITEM_ADD_WITH_MNEMONIC(menu, menuitem, _("_Save as..."), PART_MENU_SAVE_AS);
+	g_signal_connect(G_OBJECT(menuitem), "activate",
+			 G_CALLBACK(textview_part_menu_activated), textview);
+	MENUITEM_ADD_FROM_STOCK(menu, menuitem, GTK_STOCK_PRINT, PART_MENU_PRINT);
+	g_signal_connect(G_OBJECT(menuitem), "activate",
+			 G_CALLBACK(textview_part_menu_activated), textview);
+
+	MENUITEM_ADD(menu, menuitem, NULL, 0);
+
+	MENUITEM_ADD_WITH_MNEMONIC(menu, menuitem, _("_Copy file name"), PART_MENU_COPY_FILENAME);
+	g_signal_connect(G_OBJECT(menuitem), "activate",
+			 G_CALLBACK(textview_part_menu_activated), textview);
+
+	g_signal_connect(G_OBJECT(menu), "selection_done",
+			 G_CALLBACK(textview_part_menu_closed), textview);
+
+	gtk_widget_show_all(menu);
+}
+
+static void textview_add_part_widget(TextView *textview, GtkTextIter *iter,
+				     MimeInfo *mimeinfo, const gchar *str)
+{
+	GtkTextView *text = GTK_TEXT_VIEW(textview->text);
+	GtkTextBuffer *buffer;
+	GtkTextChildAnchor *anchor;
+	GtkWidget *hbox;
+	GtkWidget *ebox;
+	GtkWidget *label;
+	GtkWidget *arrow;
+	GtkStyle *style;
+	GdkColor bg = {0, 0xe000, 0xe500, 0xffff};
+	GdkColor fg = {0, 0x8000, 0x9800, 0xffff};
+	GdkColor bg2 = {0, 0xc000, 0xc800, 0xffff};
+	GdkColor fg2 = {0, 0x6000, 0x8000, 0xffff};
+
+	buffer = gtk_text_view_get_buffer(text);
+	gtk_text_buffer_insert(buffer, iter, "\n", 1);
+	anchor = gtk_text_buffer_create_child_anchor(buffer, iter);
+	ebox = gtk_event_box_new();
+	hbox = gtk_hbox_new(FALSE, 4);
+	gtk_container_set_border_width(GTK_CONTAINER(hbox), 3);
+	label = gtk_label_new(str);
+	gtk_box_pack_start(GTK_BOX(hbox), label, TRUE, TRUE, 0);
+	arrow = gtk_arrow_new(GTK_ARROW_DOWN, GTK_SHADOW_OUT);
+	gtk_box_pack_start(GTK_BOX(hbox), arrow, FALSE, FALSE, 0);
+	gtk_container_add(GTK_CONTAINER(ebox), hbox);
+	gtk_widget_show_all(ebox);
+	g_signal_connect(G_OBJECT(ebox), "button_press_event",
+			 G_CALLBACK(textview_part_widget_button_pressed),
+			 textview);
+	g_signal_connect(G_OBJECT(ebox), "enter_notify_event",
+			 G_CALLBACK(textview_part_widget_entered), textview);
+	g_signal_connect(G_OBJECT(ebox), "leave_notify_event",
+			 G_CALLBACK(textview_part_widget_left), textview);
+	g_signal_connect_after(G_OBJECT(ebox), "expose_event",
+			 G_CALLBACK(textview_part_widget_exposed), textview);
+
+	style = gtk_widget_get_style(label);
+	bg = style->bg[GTK_STATE_NORMAL];
+	fg = style->fg[GTK_STATE_NORMAL];
+	bg2 = style->bg[GTK_STATE_PRELIGHT];
+	fg2 = style->fg[GTK_STATE_PRELIGHT];
+
+	gtk_widget_modify_bg(ebox, GTK_STATE_NORMAL, &bg);
+	gtk_widget_modify_fg(ebox, GTK_STATE_NORMAL, &fg);
+	gtk_widget_modify_bg(ebox, GTK_STATE_PRELIGHT, &bg2);
+	gtk_widget_modify_fg(ebox, GTK_STATE_PRELIGHT, &fg2);
+
+	g_object_set_data(G_OBJECT(ebox), "mimeinfo", mimeinfo);
+
+	gtk_text_view_add_child_at_anchor(text, ebox, anchor);
+	gtk_text_buffer_insert(buffer, iter, "\n", 1);
+}
+
 static void textview_add_part(TextView *textview, MimeInfo *mimeinfo, FILE *fp)
 {
 	GtkTextView *text = GTK_TEXT_VIEW(textview->text);
@@ -649,9 +932,13 @@ static void textview_add_part(TextView *textview, MimeInfo *mimeinfo, FILE *fp)
 	charset = textview_get_src_encoding(textview, mimeinfo);
 
 	if (mimeinfo->mime_type == MIME_MESSAGE_RFC822) {
+		g_snprintf(buf, sizeof(buf), "%s (%s)",
+			   mimeinfo->content_type,
+			   to_human_readable(mimeinfo->content_size));
+		textview_add_part_widget(textview, &iter, mimeinfo, buf);
+		gtk_text_buffer_get_end_iter(buffer, &iter);
 		headers = textview_scan_header(textview, fp, charset);
 		if (headers) {
-			gtk_text_buffer_insert(buffer, &iter, "\n", 1);
 			textview_show_header(textview, headers);
 			procheader_header_array_destroy(headers);
 		}
@@ -666,20 +953,19 @@ static void textview_add_part(TextView *textview, MimeInfo *mimeinfo, FILE *fp)
 #endif
 
 	if (mimeinfo->filename || mimeinfo->name)
-		g_snprintf(buf, sizeof(buf), "\n[%s  %s (%s)]\n",
+		g_snprintf(buf, sizeof(buf), "%s  %s (%s)",
 			   mimeinfo->filename ? mimeinfo->filename :
 			   mimeinfo->name,
 			   mimeinfo->content_type,
 			   to_human_readable(mimeinfo->content_size));
 	else
-		g_snprintf(buf, sizeof(buf), "\n[%s (%s)]\n",
+		g_snprintf(buf, sizeof(buf), "%s (%s)",
 			   mimeinfo->content_type,
 			   to_human_readable(mimeinfo->content_size));
 
 	if (mimeinfo->mime_type != MIME_TEXT &&
 	    mimeinfo->mime_type != MIME_TEXT_HTML) {
-		gtk_text_buffer_insert_with_tags_by_name
-			(buffer, &iter, buf, -1, "mimepart", NULL);
+		textview_add_part_widget(textview, &iter, mimeinfo, buf);
 		if (mimeinfo->mime_type == MIME_IMAGE &&
 		    prefs_common.inline_image) {
 			GdkPixbuf *pixbuf;
@@ -736,8 +1022,7 @@ static void textview_add_part(TextView *textview, MimeInfo *mimeinfo, FILE *fp)
 		if (!mimeinfo->main &&
 		    mimeinfo->parent &&
 		    mimeinfo->parent->children != mimeinfo)
-			gtk_text_buffer_insert_with_tags_by_name
-				(buffer, &iter, buf, -1, "mimepart", NULL);
+			textview_add_part_widget(textview, &iter, mimeinfo, buf);
 		else
 			gtk_text_buffer_insert(buffer, &iter, "\n", 1);
 		textview_write_body(textview, mimeinfo, fp, charset);
@@ -1002,27 +1287,28 @@ static gchar *make_email_string(const gchar *bp, const gchar *ep)
 {
 	/* returns a mailto: URI; mailto: is also used to detect the
 	 * uri type later on in the button_pressed signal handler */
-	gchar *tmp;
+	gchar *tmp, *enc;
 	gchar *result;
 
 	tmp = g_strndup(bp, ep - bp);
-	result = g_strconcat("mailto:", tmp, NULL);
+	enc = uriencode_for_mailto(tmp);
+	result = g_strconcat("mailto:", enc, NULL);
+	g_free(enc);
 	g_free(tmp);
 
 	return result;
 }
 
 #define ADD_TXT_POS(bp_, ep_, pti_) \
-	if ((last->next = alloca(sizeof(struct txtpos))) != NULL) { \
-		last = last->next; \
-		last->bp = (bp_); last->ep = (ep_); last->pti = (pti_); \
-		last->next = NULL; \
-	} else { \
-		g_warning("alloc error scanning URIs\n"); \
-		gtk_text_buffer_insert_with_tags_by_name \
-			(buffer, &iter, linebuf, -1, fg_tag, NULL); \
-		return; \
-	}
+{ \
+	struct txtpos *last; \
+ \
+	last = g_new(struct txtpos, 1); \
+	last->bp = (bp_); \
+	last->ep = (ep_); \
+	last->pti = (pti_); \
+	txtpos_list = g_slist_append(txtpos_list, last); \
+}
 
 /* textview_make_clickable_parts() - colorizes clickable parts */
 static void textview_make_clickable_parts(TextView *textview,
@@ -1070,8 +1356,8 @@ static void textview_make_clickable_parts(TextView *textview,
 	struct txtpos {
 		const gchar	*bp, *ep;	/* text position */
 		gint		 pti;		/* index in parse table */
-		struct txtpos	*next;		/* next */
-	} head = {NULL, NULL, 0,  NULL}, *last = &head;
+	};
+	GSList *txtpos_list = NULL;
 
 	buffer = gtk_text_view_get_buffer(text);
 	gtk_text_buffer_get_end_iter(buffer, &iter);
@@ -1113,36 +1399,41 @@ static void textview_make_clickable_parts(TextView *textview,
 	}
 
 	/* colorize this line */
-	if (head.next) {
+	if (txtpos_list) {
 		const gchar *normal_text = linebuf;
+		GSList *cur;
 
 		/* insert URIs */
-		for (last = head.next; last != NULL;
-		     normal_text = last->ep, last = last->next) {
+		for (cur = txtpos_list; cur != NULL; cur = cur->next) {
+			struct txtpos *pos = (struct txtpos *)cur->data;
 			RemoteURI *uri;
 
 			uri = g_new(RemoteURI, 1);
-			if (last->bp - normal_text > 0)
+			if (pos->bp - normal_text > 0)
 				gtk_text_buffer_insert_with_tags_by_name
 					(buffer, &iter,
 					 normal_text,
-					 last->bp - normal_text,
+					 pos->bp - normal_text,
 					 fg_tag, NULL);
-			uri->uri = parser[last->pti].build_uri(last->bp,
-							       last->ep);
+			uri->uri = parser[pos->pti].build_uri(pos->bp, pos->ep);
 			uri->filename = NULL;
 			uri->start = gtk_text_iter_get_offset(&iter);
 			gtk_text_buffer_insert_with_tags_by_name
-				(buffer, &iter, last->bp, last->ep - last->bp,
+				(buffer, &iter, pos->bp, pos->ep - pos->bp,
 				 uri_tag, fg_tag, NULL);
 			uri->end = gtk_text_iter_get_offset(&iter);
 			textview->uri_list =
 				g_slist_append(textview->uri_list, uri);
+			normal_text = pos->ep;
+
+			g_free(pos);
 		}
 
 		if (*normal_text)
 			gtkut_text_buffer_insert_with_tag_by_name
 				(buffer, &iter, normal_text, -1, fg_tag);
+
+		g_slist_free(txtpos_list);
 	} else {
 		gtkut_text_buffer_insert_with_tag_by_name
 			(buffer, &iter, linebuf, -1, fg_tag);
@@ -1288,8 +1579,11 @@ void textview_destroy(TextView *textview)
 	clipboard = gtk_clipboard_get(GDK_SELECTION_PRIMARY);
 	gtk_text_buffer_remove_selection_clipboard(buffer, clipboard);
 
+	gtk_widget_destroy(textview->popup_menu);
+
 	textview_uri_list_remove_all(textview->uri_list);
 	textview->uri_list = NULL;
+
 	g_free(textview);
 }
 
@@ -1312,8 +1606,9 @@ void textview_set_font(TextView *textview, const gchar *codeset)
 		}
 	}
 
-	gtk_text_view_set_pixels_above_lines(text, prefs_common.line_space / 2);
+	gtk_text_view_set_pixels_above_lines(text, prefs_common.line_space - prefs_common.line_space / 2);
 	gtk_text_view_set_pixels_below_lines(text, prefs_common.line_space / 2);
+	gtk_text_view_set_pixels_inside_wrap(text, prefs_common.line_space);
 }
 
 void textview_set_position(TextView *textview, gint pos)
@@ -1344,6 +1639,27 @@ static GPtrArray *textview_scan_header(TextView *textview, FILE *fp,
 
 	return procheader_get_header_array_for_display(fp, encoding);
 }
+ 
+static void textview_size_allocate_cb(GtkWidget *widget,
+				      GtkAllocation *allocation, gpointer data)
+{
+	GtkWidget *child = GTK_WIDGET(data);
+	gint new_width;
+
+	/* g_print("textview_size_allocate_cb: (%d, %d)\n", allocation->width, allocation->height); */
+	new_width = allocation->width - 14;
+	if (new_width < -1)
+		new_width = -1;
+	gtk_widget_set_size_request(child, new_width, 16);
+}
+
+void textview_hline_destroy_cb(GtkObject *object, gpointer data)
+{
+	GtkWidget *text = GTK_WIDGET(data);
+
+	g_signal_handlers_disconnect_by_func(text, textview_size_allocate_cb,
+					     object);
+}
 
 static void textview_show_header(TextView *textview, GPtrArray *headers)
 {
@@ -1393,6 +1709,38 @@ static void textview_show_header(TextView *textview, GPtrArray *headers)
 		gtk_text_buffer_insert_with_tags_by_name
 			(buffer, &iter, "\n", 1, "header", NULL);
 	}
+
+	textview_insert_border(textview, &iter, 0);
+}
+
+static void textview_insert_border(TextView *textview, GtkTextIter *iter,
+				   gint padding)
+{
+	GtkTextView *text = GTK_TEXT_VIEW(textview->text);
+	GtkTextBuffer *buffer;
+	GtkTextChildAnchor *anchor;
+	GtkWidget *vbox;
+	GtkWidget *pad;
+	GtkWidget *hline;
+
+	buffer = gtk_text_view_get_buffer(text);
+
+	anchor = gtk_text_buffer_create_child_anchor(buffer, iter);
+	vbox = gtk_vbox_new(FALSE, 0);
+	if (padding > 0) {
+		pad = gtk_vbox_new(FALSE, 0);
+		gtk_box_pack_start(GTK_BOX(vbox), pad, FALSE, FALSE, 0);
+		gtk_widget_set_size_request(pad, -1, padding);
+	}
+	hline = gtk_hseparator_new();
+	gtk_box_pack_start(GTK_BOX(vbox), hline, FALSE, FALSE, 0);
+	gtk_widget_show_all(vbox);
+	gtk_widget_set_size_request(hline, 320, 16);
+	gtk_text_view_add_child_at_anchor(text, vbox, anchor);
+	g_signal_connect(G_OBJECT(hline), "destroy",
+			 G_CALLBACK(textview_hline_destroy_cb), text);
+	g_signal_connect(G_OBJECT(text), "size-allocate",
+			 G_CALLBACK(textview_size_allocate_cb), hline);
 }
 
 gboolean textview_search_string(TextView *textview, const gchar *str,
@@ -1809,6 +2157,10 @@ static gboolean textview_event_after(GtkWidget *widget, GdkEvent *event,
 	if (bevent->button != 1 && bevent->button != 2)
 		return FALSE;
 
+	/* don't process child widget's event */
+	if (gtk_text_view_get_window(GTK_TEXT_VIEW(widget), GTK_TEXT_WINDOW_TEXT) != bevent->window)
+		return FALSE;
+
 	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(widget));
 
 	/* don't follow a link if the user has selected something */
@@ -1817,7 +2169,7 @@ static gboolean textview_event_after(GtkWidget *widget, GdkEvent *event,
 		return FALSE;
 
 	gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(widget),
-					      GTK_TEXT_WINDOW_WIDGET,
+					      GTK_TEXT_WINDOW_TEXT,
 					      bevent->x, bevent->y, &x, &y);
 	gtk_text_view_get_iter_at_location(GTK_TEXT_VIEW(widget), &iter, x, y);
 	on_link = textview_get_link_tag_bounds(textview, &iter, &start, &end);
@@ -1825,7 +2177,7 @@ static gboolean textview_event_after(GtkWidget *widget, GdkEvent *event,
 		return FALSE;
 
 	uri = textview_get_uri(textview, &start, &end);
-	if (!uri)
+	if (!uri || !uri->uri)
 		return FALSE;
 
 	if (!g_ascii_strncasecmp(uri->uri, "mailto:", 7)) {
@@ -1837,6 +2189,8 @@ static gboolean textview_event_after(GtkWidget *widget, GdkEvent *event,
 		if (ac && ac->protocol == A_NNTP)
 			ac = NULL;
 		compose_new(ac, msginfo->folder, uri->uri + 7, NULL);
+	} else if (uri->uri[0] == '#') {
+		/* don't open in-page link */
 	} else if (textview_uri_security_check(textview, uri) == TRUE)
 		open_uri(uri->uri, prefs_common.uri_cmd);
 
@@ -1917,7 +2271,9 @@ static gboolean textview_motion_notify(GtkWidget *widget,
 				       TextView *textview)
 {
 	gint x, y;
-
+ 
+	if (gtk_text_view_get_window(GTK_TEXT_VIEW(widget), GTK_TEXT_WINDOW_TEXT) != event->window)
+		return FALSE;
 	gtk_text_view_window_to_buffer_coords(GTK_TEXT_VIEW(widget),
 					      GTK_TEXT_WINDOW_WIDGET,
 					      event->x, event->y, &x, &y);
@@ -1968,6 +2324,14 @@ static gboolean textview_visibility_notify(GtkWidget *widget,
 	gtk_widget_show(menuitem);					\
 }
 
+#define ADD_MENU_SEPARATOR()						\
+{									\
+	menuitem = gtk_menu_item_new();					\
+	gtk_widget_set_sensitive(menuitem, FALSE);			\
+	gtk_menu_shell_append(GTK_MENU_SHELL(menu), menuitem);		\
+	gtk_widget_show(menuitem);					\
+}
+
 static void textview_populate_popup(GtkWidget *widget, GtkMenu *menu,
 				    TextView *textview)
 {
@@ -1978,6 +2342,7 @@ static void textview_populate_popup(GtkWidget *widget, GtkMenu *menu,
 	gboolean on_link;
 	RemoteURI *uri;
 	GdkPixbuf *pixbuf;
+	gchar *selected_text;
 
 	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(widget));
 
@@ -1998,13 +2363,17 @@ static void textview_populate_popup(GtkWidget *widget, GtkMenu *menu,
 		ADD_MENU(_("Sa_ve this image as..."),
 			 textview_popup_menu_activate_image_cb);
 	}
+
+	selected_text = gtkut_text_view_get_selection(GTK_TEXT_VIEW(widget));
+
+	uri = NULL;
 	on_link = textview_get_link_tag_bounds(textview, &iter, &start, &end);
 	if (!on_link)
-		return;
+		goto finish;
 
 	uri = textview_get_uri(textview, &start, &end);
-	if (!uri)
-		return;
+	if (!uri || !uri->uri)
+		goto finish;
 
 	separator = gtk_separator_menu_item_new();
 	gtk_menu_shell_append(GTK_MENU_SHELL(menu), separator);
@@ -2013,6 +2382,9 @@ static void textview_populate_popup(GtkWidget *widget, GtkMenu *menu,
 	if (!g_ascii_strncasecmp(uri->uri, "mailto:", 7)) {
 		ADD_MENU(_("Compose _new message"),
 			 textview_popup_menu_activate_open_uri_cb);
+		ADD_MENU(_("R_eply to this address"),
+			 textview_popup_menu_activate_reply_cb);
+		ADD_MENU_SEPARATOR();
 		ADD_MENU(_("Add to address _book..."),
 			 textview_popup_menu_activate_add_address_cb);
 		ADD_MENU(_("Copy this add_ress"),
@@ -2023,6 +2395,14 @@ static void textview_populate_popup(GtkWidget *widget, GtkMenu *menu,
 		ADD_MENU(_("Copy this _link"),
 			 textview_popup_menu_activate_copy_cb);
 	}
+
+finish:
+	syl_plugin_signal_emit("textview-menu-popup", menu,
+			       GTK_TEXT_VIEW(widget),
+			       uri ? uri->uri : NULL,
+			       selected_text,
+			       textview->messageview->msginfo);
+	g_free(selected_text);
 }
 
 static void textview_popup_menu_activate_open_uri_cb(GtkMenuItem *menuitem,
@@ -2048,34 +2428,86 @@ static void textview_popup_menu_activate_open_uri_cb(GtkMenuItem *menuitem,
 		if (ac && ac->protocol == A_NNTP)
 			ac = NULL;
 		compose_new(ac, msginfo->folder, uri->uri + 7, NULL);
+	} else if (uri->uri[0] == '#') {
+		/* don't open in-page link */
 	} else if (textview_uri_security_check(textview, uri) == TRUE)
 		open_uri(uri->uri, prefs_common.uri_cmd);
 }
 
-static void textview_popup_menu_activate_add_address_cb(GtkMenuItem *menuitem,
-							gpointer data)
+static void textview_popup_menu_activate_reply_cb(GtkMenuItem *menuitem,
+						  gpointer data)
 {
 	RemoteURI *uri = (RemoteURI *)data;
-	const gchar *addr;
+	TextView *textview;
 
 	g_return_if_fail(uri != NULL);
 
 	if (!uri->uri)
 		return;
 
-	if (!g_ascii_strncasecmp(uri->uri, "mailto:", 7))
-		addr = uri->uri + 7;
-	else
-		addr = uri->uri;
+	textview = g_object_get_data(G_OBJECT(menuitem), "textview");
+	g_return_if_fail(textview != NULL);
+
+	if (!g_ascii_strncasecmp(uri->uri, "mailto:", 7)) {
+		MsgInfo *msginfo = textview->messageview->msginfo;
+		ComposeMode mode = COMPOSE_REPLY;
+		gchar *text;
+		GList *list;
+		Compose *compose;
+
+		g_return_if_fail(msginfo != NULL);
+
+		if (prefs_common.reply_with_quote)
+			mode |= COMPOSE_WITH_QUOTE;
+
+		text = gtkut_text_view_get_selection
+			(GTK_TEXT_VIEW(textview->text));
+		if (text && *text == '\0') {
+			g_free(text);
+			text = NULL;
+		}
+
+		compose_reply(msginfo, msginfo->folder, mode, text);
+		list = compose_get_compose_list();
+		list = g_list_last(list);
+		if (list && list->data) {
+			compose = (Compose *)list->data;
+			compose_block_modified(compose);
+			compose_entry_set(compose, uri->uri + 7,
+					  COMPOSE_ENTRY_TO);
+			compose_unblock_modified(compose);
+		}
+		g_free(text);
+	}
+}
+
+static void textview_popup_menu_activate_add_address_cb(GtkMenuItem *menuitem,
+							gpointer data)
+{
+	RemoteURI *uri = (RemoteURI *)data;
+	gchar *addr;
+
+	g_return_if_fail(uri != NULL);
+
+	if (!uri->uri)
+		return;
+
+	if (!g_ascii_strncasecmp(uri->uri, "mailto:", 7)) {
+		addr = g_malloc(strlen(uri->uri + 7) + 1);
+		decode_uri(addr, uri->uri + 7);
+	} else
+		addr = g_strdup(uri->uri);
 
 	addressbook_add_contact(addr, addr, NULL);
+
+	g_free(addr);
 }
 
 static void textview_popup_menu_activate_copy_cb(GtkMenuItem *menuitem,
 						 gpointer data)
 {
 	RemoteURI *uri = (RemoteURI *)data;
-	const gchar *uri_string;
+	gchar *uri_string;
 	GtkClipboard *clipboard;
 
 	g_return_if_fail(uri != NULL);
@@ -2083,15 +2515,18 @@ static void textview_popup_menu_activate_copy_cb(GtkMenuItem *menuitem,
 	if (!uri->uri)
 		return;
 
-	if (!g_ascii_strncasecmp(uri->uri, "mailto:", 7))
-		uri_string = uri->uri + 7;
-	else
-		uri_string = uri->uri;
+	if (!g_ascii_strncasecmp(uri->uri, "mailto:", 7)) {
+		uri_string = g_malloc(strlen(uri->uri + 7) + 1);
+		decode_uri(uri_string, uri->uri + 7);
+	} else
+		uri_string = g_strdup(uri->uri);
 
 	clipboard = gtk_clipboard_get(GDK_SELECTION_PRIMARY);
 	gtk_clipboard_set_text(clipboard, uri_string, -1);
 	clipboard = gtk_clipboard_get(GDK_SELECTION_CLIPBOARD);
 	gtk_clipboard_set_text(clipboard, uri_string, -1);
+
+	g_free(uri_string);
 }
 
 static void textview_popup_menu_activate_image_cb(GtkMenuItem *menuitem,
